"""
Hakku - tool for article clustering based on word counts in titles
and keywords. 

Clustering can be run from inside a script by run_files function or
from command line with python -m hakku articles1.txt articles2.txt
"""

import sys
from codecs import open as copen
import numpy as np

from nltk.stem.snowball import EnglishStemmer

from .article import get_articles, Source
from .word_occurrance import append_from_file, create_article_word_matrix,\
        create_article_author_matrix, ask_for_source, ask_for_stopwords
from .analysis import analysis, create_csv
from .visualization import visualize

def run_files(fnames, outfile="result.csv", low_bound=10, up_bound=20):
    """
    fnames      --  Array filenames for files containing the article data
    outfile     --  Filename for clustering output
    low_bound   --  criteria to stop clustering iteration of certain cluster
    up_bound    --  criteria to stop clustering

    Do cluster analysis for articles contained in files listed in 
    fnames array.

    Result of the clustering is stored in csv file.

    Addittionally user can also select to visualize the clustering hierarchy.
    """

    if not fnames:
        print_help()
        return

    articles = []
    for fname in fnames:
        with copen(fname, 'r', encoding='utf-8') as f:
            articles += get_articles(f, ask_for_source(fname))

    # some default stopwords
    extended_stopwords = ["elsevier", "copyright"]
    extended_stopwords += ["based", "paper"]
    stemmed_extended_stopwords = ["studi", "use", "analysi", "method", "differ",
            "result", "model", "also", "signific", "among", "within"]

    EXTENDED_STOPW_FILENAME = 'stopwords_whole.txt'
    EXTENDED_STOPSW_FILENAME = 'stopwords_stem.txt'
    #extended_stopwords = append_from_file(EXTENDED_STOPWORDS,
                                          #EXTENDED_STOPW_FILENAME,
                                          #'Only default stopwords used!')
    #stemmed_extended_stopwords = append_from_file(STEMMED_EXTENDED_STOPWORDS,
                                                  #EXTENDED_STOPSW_FILENAME,
                                                  #'Stemmed stopwords not used!')

    words = []

    try:
        f = open("last_stopwords.txt", "r")
        lines = f.readlines()
        lines = [l[:-1] for l in lines]
        print("Previous stopwords:")
        print(" ".join(lines))
        ans = input("Use previous stopwords [y/N]: ")[0]
        if ans == "y" or ans == "Y":
            words += lines

        f.close()
    except OSError:
        pass
    except IOError:
        pass

    # TODO ask for stemmed words also
    glob_st = EnglishStemmer()
    words += ask_for_stopwords()
    words_stemmed = map(glob_st.stem, words)
    extended_stopwords += words
    stemmed_extended_stopwords += words_stemmed

    print("Used stopwords: " + ", ".join(words))

    try:
        f = open("last_stopwords.txt", "w")
        f.writelines([l + "\n" for l in words])
    except OSError:
        pass

    print("Starting clustering...")

    ### Create word occurrence matrix.
    (m, ai, words) = create_article_word_matrix(articles,
            extended_stopwords ,stemmed_extended_stopwords)

    ### Create author occurrence matrix.
    (m_authors, uniq_authors) = create_article_author_matrix(articles)

    # Create dictionary structure and save for further analysis in a file.
    data = {}
    data['m'] = np.array(m)
    data['ai'] = np.array(ai)
    data['year'] = np.array([a.year for a in articles])
    data['cite_count'] = np.array([a.cite_count for a in articles])
    data['words'] = np.array(words)
    data['m_authors'] = np.array(m_authors)
    data['uniq_authors'] = np.array(uniq_authors)
    data['articles'] = np.array(articles)

    # Print the whole wordlist to output file:
    wfreq = np.sum(m, 0).astype(int).tolist()
    pairs = zip(wfreq, words)
    kfunc = lambda x: x[0]
    pairs = reversed(sorted(pairs, key=kfunc))
    with open('words_frequency.out','w') as f:
        for (fr, w) in pairs:
            f.write('%-7d %s\n'%(fr, w))

    # perform clustering
    c, names, inds = analysis(data['m'], data['words'], data['ai'], up_bound, low_bound)

    # save results to csv
    create_csv(c, names, data['cite_count'][inds],
            data['articles'][inds], data['year'][inds], outfile)

    # create tree visualization
    ans = input("Visualize: [y/N]: ")
    ans = ans[0]

    if ans == "y" or ans == "Y":
        visualize(outfile)

def print_help():
    helptext = """usage: python -m hakku articles1.txt [articles2.txt ...]"""

    print(helptext)




#!/usr/bin/python
# -*- coding: utf-8 -*-
import sys, csv
import codecs
import re
from enum import Enum
import csv

class Source(Enum):
    WebOfKnowledge = 1
    IEEE = 2
    ScienceDirect = 3

class Article:
    """Article. Reads from string to an object."""

    def __init__(self, source=u"", title=u"", doi=u""):
        """Creates an article."""
        self.title = title
        self.authors = []
        self.abstract = u""
        self.cite_count = 0
        self.doi = doi
        self.source = source
        self.source_type = u""
        self.volume = u""
        self.issue = u""
        self.page_begin = u""
        self.page_end = u""
        # self.year = 0 <-- Missing. What is a safe default? 
        self.year = 1970 # is this sensible?
        self.issn = u""

class PeekingLineReader:
    """Reads a file line-by-line, peeking into the next line. Assume that
       the file is opened as unicode using codecs.open. TODO: Handling
       of end-of-file is not implemented!
    """
    def __init__(self, ifile):
        self.fh = ifile
        self.buffer_line()

    def buffer_line(self):
        self.next_line = self.fh.readline()

    def peek2(self):
        """Peek ahead to check the next two characters."""
        return self.next_line[0:2]

    def readline(self):
        tmp = self.next_line
        self.buffer_line()
        return tmp;

    def skip_empty_rows(self):
        """Skips whitespace lines. Next readline() will return non-white."""
        while self.next_line.strip() == '':
            self.buffer_line()

def parse_string(fh):
    """Parse string field."""
    result = u''
    line = fh.readline().encode('utf-8')
    line = line[3:].rstrip()
    result = line.decode("utf-8")
    while fh.peek2() == '  ':
        result = result + ' '
        line = fh.readline().encode('utf-8')
        line = line[3:].rstrip()
        result = result + line.decode("utf-8")
    return result

def parse_string_list(fh):
    """Parse list of strings."""
    result = []
    line = fh.readline().encode('utf-8')
    line = line[3:].rstrip()
    result.append(line)
    while fh.peek2() == '  ':
        line = fh.readline().encode('utf-8')
        line = line[3:].rstrip()
        result.append(line)
    return result

def parse_int(fh):
    """Parse int field."""
    line = fh.readline().encode('utf-8')
    line = line[3:].rstrip()
    return int(line)


def parse_article_wok(fh):
    """Parses one article from Web of Knowldedge database. 
    Continues until ER code occurs. 
    http://images.webofknowledge.com/WOKRS57B4/help/WOK/hs_alldb_fieldtags.html
    Input: fh - PeekingLineReader instance."""

    article = Article()

    while True:
        code = fh.peek2()
        if code == u'ER':
            fh.readline()
            break
        elif code == u'TI':
            article.title = parse_string(fh)
        elif code == u'AU':
            article.authors = parse_string_list(fh)
        elif code == u'CA':
            article.authors.append(parse_string(fh))
        elif code == u'AB':
            article.abstract = parse_string(fh)
        elif code == u'Z9':
            article.cite_count = parse_int(fh)
        elif code == u'DI':
            article.doi = parse_string(fh)
        elif code == u'PT':
            article.source_type = parse_string(fh)
        elif code == u'SO':
            article.source = parse_string(fh)
        elif code == u'VL':
            article.volume = parse_string(fh)
        elif code == u'IS':
            article.issue = parse_string(fh)
        elif code == u'BP':
            article.page_begin = parse_string(fh)
        elif code == u'EP':
            article.page_end = parse_string(fh)
        elif code == u'PY':
            article.year = parse_int(fh)
        elif code == u'SN':
            article.issn = parse_string(fh)
        else:
            fh.readline()

    return article

def parse_articles_ieee(f):
    creader = csv.reader(f, delimiter=",")
    header = False
    articles = []
    for row in creader:
        if not header:
            header = True
            continue
        article = Article()
        article.title = row[0]
        article.authors = row[1]
        article.issue = row[3]
        article.year = int(row[5])
        article.volume = row[6]
        article.issue = row[7]
        article.page_begin = row[8]
        article.page_end = row[9]
        article.abstract = row[10]
        article.issn = row[11]
        article.doi = row[13]
        if row[21]:
            article.cite_count = int(row[21])
        
        articles.append(article)
    return articles

def parse_article_sd(f):
    article = Article()
    line = f.readline()[:-1]

    if not line:
        line = f.readline()[:-1]

    if not line:
        return None

    while line:
        letters = line[0:2]
        rest = line[6:].strip(" \t\n\r")
        if letters == "TY":
            article.source_type = rest[0]
        elif letters == "T1" or letters == "TI":
            article.title = rest
        elif letters == "JO":
            article.source = rest
        elif letters == "VL":
            article.volume = rest
        elif letters == "IS":
            article.issue = rest
        elif letters == "SP":
            article.page_begin = rest
        elif letters == "EP":
            article.page_end = rest
        elif letters == "PY":
            year_str = rest.split("/")[0]
            if len(year_str) == 4:
                article.year = int(rest[0:4])
        elif letters == "AU":
            article.authors.append(rest)
        elif letters == "SN":
            article.issn = rest
        elif letters == "DO":
            article.doi = rest
        elif letters == "AB":
            abstract = f.readline()
            article.abstract = abstract[:-1].strip(" \t\n\r")
        elif letters == "ER":
            break

        line = f.readline()[:-1]

    return article

def get_articles(f, source):
    """Gives a list of articles."""

    articles = []

    if source == Source.WebOfKnowledge:
        fh = PeekingLineReader(f)

        while True:
            articles.append(parse_article_wok(fh))
            fh.skip_empty_rows()
            if fh.peek2() == 'EF':
                break

    elif source == Source.IEEE:
        articles = parse_articles_ieee(f)

    elif source == Source.ScienceDirect:
        a = parse_article_sd(f)
        while a:
            articles.append(a)
            a = parse_article_sd(f)

    return articles

if __name__ == "__main__":
    articles = get_articles(sys.stdin)





# -*- coding: utf-8 -*-
""" Heuristic synonym detector and combiner. """

import string, re

CONVERSIONS = {'optimisation':'optimization',
               'maximisation':'maximization',
               'alexandroff':'alexandrov',
               'kullback-liebler':'kullback-leibler',
               'modelling':'modeling',
               'labelling':'labeling',
               '-based':' based'}

def process_word(word):
    word = re.sub('  *',' ',word)
    word = re.sub('--*','-',word)
    word = re.sub(' - ','-',word)
    word = re.sub(' \(.*?\)$','',word)

    # FIXME: This is not a good way to handle plurals:
    word = re.sub('sses$','ss',word)
    word = re.sub('(?<!ser)ies$','y',word)
    word = re.sub('(?<!s)s$','',word)

    # Revert some broken words:
    word = re.sub('qo$','qos',word)
    word = re.sub('chao$','chaos',word)
    word = re.sub('baye$','bayes',word)

    word = re.sub('kurtosi$','kurtosis',word)
    word = re.sub('analysi$','analysis',word)
    word = re.sub('basi$','basis',word)
    word = re.sub('genesi$','genesis',word)
    word = re.sub('synthesi$','synthesis',word)
    word = re.sub('axi$','axis',word)
    word = re.sub('hypothesi$','hypothesis',word)
    word = re.sub('diagnosi$','diagnosis',word)
    word = re.sub('metropoli$','metropolis',word)

    word = re.sub('consensu$','consensus',word)
    word = re.sub('bia$','bias',word)
    word = re.sub('focu$','focus',word)
    word = re.sub('campu$','campus',word)
    word = re.sub('heterogeneou$','heterogeneous',word)
    word = re.sub('continuou$','continuous',word)
    word = re.sub('viru$','virus',word)    
    word = re.sub('versu$','versus',word)

    word = re.sub('meshe$','mesh',word)
    
    word = re.sub('k-mean$','k-means',word)
    word = re.sub('knows what it know$','knows what it knows',word)
    word = re.sub('least square$','least squares',word)

    for a,b in CONVERSIONS.iteritems():
        word = string.replace(word,a,b)
        
    return word

def detect_synonyms(uniqlist):
    """Returns a reduced word list, and a dictionary that maps
    original words to indices of reduced words.

    Note: uniqlist must contain each word only once.
    """

    ind_of_word = {}
    reduced_list = []

    for original_word in uniqlist:
        # apply conversions and heuristics
        normalized_word = process_word(original_word)
        if reduced_list.count(normalized_word)>0:
            index = reduced_list.index(normalized_word)
        else:
            index = len(reduced_list)
            reduced_list.append(normalized_word)
            
        ind_of_word[original_word] = index

    return (reduced_list, ind_of_word)




import sys
import numpy as np
import pkg_resources as pkgr


def make_newick(tree_dir, _r=True):
    """
    Generate newick recursively from dict tree

    _r is recursion parameter, which should be ignored
    """

    if len(tree_dir) <= 0:
        # empty tree
        return ""

    s = "("
    comma = ""
    # loop cluster ids as node names
    for key in tree_dir.keys():
        # generate newick for subtree
        r = make_newick(tree_dir[key][0], _r=False)
        # add subtree and node name
        s += comma + r + str(key)
        comma = ","

    s += ")"

    # if it is level 1, add semicolon
    # to end newick tree
    if _r:
        s += ";"

    return s


def visualize(filename, delimiter='\t'):
    try:
        from ete3 import Tree, TreeStyle, TextFace
    except ImportError:
        print("ete3 not found, install with extras to use visualization")
        return

    # first line of the csv is header so skip it
    headerread = False

    # dir for tree structure
    ctree = {}

    with open(filename, 'r') as f:
        for line in f:
            if not headerread:
                # skip header
                headerread = True
            else:
                split = line.split(delimiter)

                # cluster ids
                first = split[0]
                clusters = first.split()

                # citation count
                cit = float(split[8])

                # cluster names
                names = [""]
                names += split[1:5]

                t = ctree

                # traverse the tree with the ids
                # add any nodes that are missing
                for i in range(len(clusters)):
                    # get cluster
                    k = clusters[i]
                    if k in t:
                        # cluster id in tree, add count
                        t[k] = (t[k][0], t[k][1]+1, t[k][2]+cit, names[i])
                    else:
                        # cluster missing from tree, add new node
                        t[k] = ({}, 1, cit, names[i])
                    # move to next node
                    t = t[k][0]

    # get newick string and create tree object
    # use format = 1 to keep node names
    tree_str = make_newick(ctree)
    tree = Tree(tree_str, format=1)

    # find size of biggest article cluster
    largest_cluster = ctree[list(ctree.keys())[0]][1]
    # find number of citations in the most cited cluster
    citmax = ctree[list(ctree.keys())[0]][2]
    if citmax < 1:
        citmax = 1

    # scale parameter to size cluster representation circles
    scale = 200 / np.sqrt(largest_cluster / np.pi)

    # get colormap for citation coloring
    viridis = np.load(pkgr.resource_filename("hakku", "viridis.npy"))

    # node layout func
    def my_layout(node):
        # get the node cluster ids from node names
        ks = list(reversed([n.name for n in node.get_ancestors()]))[1:]

        # visulization root node name is "", skip it
        if node.name:
            # cluster id list doesnt contain
            # current nodes name
            ks.append(node.name)

            # cluster size
            c_size = 0
            # citation count
            c_cit = 0
            # cluster name
            c_name = ""

            # get current nodes size and cit count
            # with cluster id list
            c = ctree
            for key in ks:
                c_size = c[key][1]
                c_cit = c[key][2]
                c_name = c[key][3]
                c = c[key][0]

            # use circle to represent cluster size
            node.img_style["shape"] = "circle"
            # circle area represents size
            circle_size = scale * np.sqrt(c_size / np.pi)
            node.img_style["size"] = circle_size

            # color with total citation count
            col = get_from_colormap(viridis, c_cit / citmax)
            node.img_style["fgcolor"] = color_to_hex(
                int(255*col[0]),
                int(255*col[1]),
                int(255*col[2]))

            # cluster size as number inside the circle
            fsize = 18
            # if circle is small use small font
            if circle_size < fsize * 2:
                fsize = 11
            f1 = TextFace(str(c_size), fsize=fsize, fgcolor="white", bold=True)
            # move text to center of the circle
            f1.margin_left = -circle_size / 2. \
                - f1.get_bounding_rect().width() / 2.
            node.add_face(f1, 0, position="branch-right")

            # show cluster ids
            f2 = TextFace(" ".join(ks))
            node.add_face(f2, 0, position="branch-bottom")

            # show cluster name
            f3 = TextFace(c_name)
            node.add_face(f3, 0, position="branch-top")

    # set tree style with custom layout
    s = TreeStyle()
    s.layout_fn = my_layout
    s.show_leaf_name = False
    # add extra space
    s.branch_vertical_margin = 10
    tree.show(tree_style=s)


def color_to_hex(r, g, b):
    return "#" + hex(
        (b % 256) * 256**0
        + (g % 256) * 256**1
        + (r % 256) * 256**2)[2:]


def get_from_colormap(cmap, val):
    return cmap[int((len(cmap)-1) * val)]


if __name__ == "__main__":
    if len(sys.argv) > 2:
        visualize(sys.argv[1], delimiter=sys.argv[2])
    else:
        visualize(sys.argv[1])



"""
Functions for cluster analysis and result saving
"""

import numpy as np
from scipy.spatial.distance import pdist, squareform
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.sparse.linalg import svds
from scipy.stats import mode

THIS_YEAR = 2014

def analysis(m, words, ai, up_bound, low_bound):
    """
    m           --  articles x words matrix
    words       --  list of words correspoding to m's columns
    ai          --  article/journal list
    up_bound    --  criteria to stop clustering
    low_bound   --  criteria to stop clustering iteration of certain cluster

    returns     --  (cluster ids, cluster names) for articles

    Example:
    analysis(m, words, articles, ai, cite_count, year, 25, 10)

    Stops when largest cluster residual is 25 and don't cluster
    again cluster which size is less than 10.

    """

    print(m.shape)

    # select only feature words that occur more than once
    good_features = np.sum(m, axis=0) > 1
    # TODO filter short words
    flen = lambda x: len(x) < 3
    short_inds = np.array(list(map(flen, words)))
    good_features[short_inds] = False
    m = m[:, good_features]
    words = np.array([str(w).strip() for w in words])
    words = words[good_features]

    # dismiss articles that have no words included in clustering
    good_articles = np.sum(m, axis=1) > 0
    m = m[good_articles, :]
    ai = ai[good_articles]

    print(m.shape)

    raja = m.shape[0]

    # clustering roind index, 1 == initial unclustered
    kk = 1

    c = np.ones(shape=(raja, 1))

    # next clustering rounds article words matrix
    m_clust = m
    # corresponding word list
    words_cl = [words]

    # indix matrix for cluster names
    # words_cl[n][inds[m,n+1]] = cluster n's name for article m
    inds = np.zeros(shape=(m.shape[0], 6), dtype=np.int32)

    # boundary for the largest "redundant" cluster
    while raja > up_bound:
        print("Clustering level: " + str(kk))

        kk += 1
        if kk == 6:
            break

        # for cluster id numbering
        max_c_plus = 0

        # add new column to cluster ids
        if kk >= c.shape[1]:
            c_new = np.ones(shape=(c.shape[0], kk))
            c_new[:, 0:c.shape[1]] = c
            c = c_new
        c[:, kk-1] = 0
        
        # find unique clusters
        c_uni = np.unique(c[:, kk-2])

        # do clustering to each of last rounds clusters
        for ik in range(0, len(c_uni)):
            # indices for current cluster
            idx = c_uni[ik] == c[:, kk-2]

            # cc = cluster ids for articles
            # ii = cluster names for articles
            cc, ii = cluster_articles(m_clust[idx, :], low_bound)
            c[idx, kk-1] = cc.squeeze()
            inds[idx, kk-1] = ii.squeeze()

            # maintain sequential numbering of all the found clusters
            # if clustering results in ids 1,2,3
            # => add 3 to all coming cluster ids
            c[idx, kk-1] += max_c_plus
            max_c_plus = np.max(np.unique(c[:, kk-1]), axis=0)

        # if there are features, which are not used we should remove those
        if np.min(np.unique(inds[:, kk-1])) < 0:
            kkix = inds[:, kk-1] == 0
            inds[kkix, kk-1] = inds[kkix, kk-2]

        # remove cluster names (most common words)
        # from clustering in next level
        to_remove = np.unique(inds[:,kk-1])
        to_remove_i = np.ones((m_clust.shape[1],), dtype=bool)
        to_remove_i[to_remove] = False
        words_cl.append(words_cl[kk-2][to_remove_i])
        m_clust = m_clust[:,to_remove_i]

        # largest redundant cluster (??)
        raja = mode(c[:, kk-1])[1][0]

    # return cluster ids and cluster names
    return c, np.array([w[inds[:,i+1]] for i,w in enumerate(words_cl[:-1])]), good_articles

def create_csv(c, names, cite_count, articles, year, fname):
    """
    Creates a csv file of the clustering results

    c           --  clusters from analysis()
    names       --  cluster names from analysis()
    cite_count  --  Number of citations per article
    year        --  Articles years
    fname       --  Output file name
    """

    # delimiter to use in csv
    # note: ',' is bad since abstracts contain commas.
    delimiter = "\t"

    # create header row
    header = []
    header.append('\'Cluster ID\'')
    header.append('\'Cl. name\'')
    header.append('\'Cl. name\'')
    header.append('\'Cl. name\'')
    header.append('\'Cl. name\'')
    header.append('\'Cl. mean citation\'')
    header.append('\'Cl. mean weighted citation\'')
    header.append('\'Cl. median weighted citation\'')
    header.append('\'Article total citations\'')
    header.append('\'Article title\'')
    header.append('\'Journal title\'')
    header.append('\'Volume\'')
    header.append('\'DOI\'')
    header.append('\'ISSN\'')
    header.append('\'Page begin\'')
    header.append('\'Page end\'')
    header.append('\'Journal/Series/etc.\'')
    header.append('\'Abstract\'')

    # cluster mean citation
    c_cite = np.zeros(shape=(c.shape[0],))
    # citation mean, weighted with publising year
    c_citeW = np.zeros(shape=(c.shape[0],))
    # median citations, weighted with publising year
    c_citemW = np.zeros(shape=(c.shape[0],))

    # old versions of numpy didn't have axis in unique
    if np.__version__ < '1.13':
        c_uni = np.vstack({tuple(row) for row in c})
    else:
        c_uni = np.unique(c, axis=0)

    # calculate means and median
    for i in range(c_uni.shape[0]):
        inds = [np.all(c[j] == c_uni[i]) for j in range(c.shape[0])]
        m = np.mean(cite_count[inds])
        if np.all(THIS_YEAR - year[inds] > 0):
            w = np.mean(cite_count[inds] / (THIS_YEAR - year[inds]))
            mw = np.median(cite_count[inds] / (THIS_YEAR - year[inds]))
        else:
            w = np.mean(cite_count[inds])
            mw = np.median(cite_count[inds])
        c_cite[inds] = m
        c_citeW[inds] = w
        c_citemW[inds] = mw

    # generate csv rows
    rows = []

    for i in range(c.shape[0]):
        row = []
        # cluster ids eg. 1 2 3 4 5
        row.append(" ".join(c[i].astype(int).astype(str)))
        # cluster name stems eg. classroom cleaner click clinician
        # currently broken
        for j in range(len(names)):
            row.append(names[j][i])

        # append citation statistics
        row.append("'" + str(c_cite[i]) + "'")
        row.append("'" + str(c_citeW[i]) + "'")
        row.append("'" + str(c_citemW[i]) + "'")

        # article citations
        row.append(str(cite_count[i]))

        # add article info
        # replace 's with "s, for excel
        row.append("'" + str(articles[i].title).replace("'", "\"") + "'")
        row.append(str(articles[i].source).replace("'", "\""))
        row.append(str(articles[i].volume))
        row.append(str(articles[i].doi))
        row.append(str(articles[i].issn))
        row.append(str(articles[i].page_begin))
        row.append(str(articles[i].page_end))
        row.append(str(articles[i].source_type))
        row.append("'" + str(articles[i].abstract) + "'\n")

        # join to a single string
        rows.append(delimiter.join(row))

    # write to file
    fil = open(fname, "w")
    fil.write(delimiter.join(header) + "\n")
    fil.writelines(rows)

def diffusion_map(x, epsilon, kernel='euclidean', a=0, p=2, t=1):
    """
    Creates a dimension-reduced version of input matrix x.
    Compares all the time points to each other and maps
    using only the eigenvalues and eigenvectors.
    Needs pdist (in Octave statistics package or Matlab statistics toolbox).
    2010 Tuomo Sipola (tuomo.sipola@jyu.fi)
      [V,l,K,S] = diffusion_map(x, epsilon, kernel, a, p, t)
    Parameters:
      x       Input data which should be in format time x parameters.
      epsilon Epsilon value for the kernel.
      kernel  Can be 'euclidean' (default) or 'hamming'. Optional.
      a       The alpha value for diffusion families, 0, 0.5 or 1. Optional.
      p       To which power the distance is put.
      t       How many time steps to use, default 1. Optional.
    Returns:
      V       Eigenvectors in format values x vectors.
      l       Eigenvalues in descending order.
      K       Weight matrix of the graph.
      S       The symmetric graph Laplacian, S^t is good for clustering.

    kernel_params=''
    if kernel == 'weighted_hamming':
        kernel = wh_pdist
        kernel_params = number_of_categories(x)

    if kernel == 'bd'quareform
        disp('Using binary similarity')
        kernel = bs_pdist
        kernel_params = np.max(np.sum(x, axis=1))
        """

    # Create the kernel.
    K = squareform(pdist(x, kernel))
    K = np.exp(- K ** p / epsilon)
    # Create different families of diffusions.
    if a > 0:
        Q = np.diag(np.sum(K, axis=0))
        W = np.dot(np.dot(Q ** - a, K), Q ** - a)
    else:
        W = np.copy(K)

    # Calculate the integral of weights.
    d = np.sum(W, axis=0)
    # Create a Markov probability matrix and symmetrize it.
#D2 = D^-0.5;
    D2 = np.diag(1.0 / np.sqrt(d))
    S = np.dot(np.dot(D2, W), D2)
    # Calculate the 32 first eigenvalues and eigenvectors. (svds(S, 32))
    U, l, _ = svds(S, min(32, min(S.shape)-1))
    #l = np.diag(l ** t)
    l = l**t
    # Take right eigenvectors and normalize with the constant first eigenvector.
    #V = np.dot(D2,U)
    V = np.matmul(D2, U)
    V = V / V[0, 0]

    return V, l

def map_coordinates(V, l, N):
    Psi = np.matmul(V, np.diag(l))
    return Psi[:, N]

def silhouette(X, c):
    """Implementation
    of matlabs silhouette function

    The silhouette value for each point is a measure of
    how similar that point is to points in its own cluster,
    when compared to points in other clusters. The
    silhouette value for the ith point, Si, is defined as
    Si = (bi-ai)/ max(ai, bi)
    where ai is the average distance from the ith point to
    the other points in the same cluster as i, and bi is the
    minimum average distance from the ith point to points
    in a different cluster, minimized over clusters.
    """
    D = squareform(pdist(X))
    N = X.shape[0]

    k = len(np.unique(c))
    inds = [i+1 == c for i in range(k)]

    a = np.zeros(N)
    b = np.zeros(N)
    for i in range(N):
        # same cluster inds
        kk = inds[c[i]-1].copy()
        # dont count the studied point to mean
        kk[i] = False

        # mean distance to same cluster
        a[i] = np.mean(D[i][kk])

        # minimum mean distance to other clusters
        b[i] = np.min([np.mean(D[i][inds[j]]) for j in range(k) if j != i])

    return (b-a) / np.maximum(a, b)

def cluster_articles(m, low_b):
    """
    Cluster data points
    m       --  articles x words matrix
    low_b   --  don't do clustering for less than low_b articles
    """

    if m.shape[0] <= low_b:
        c = np.ones(shape=(m.shape[0], 1))
        inds = np.zeros(shape=(c.shape), dtype=np.int32)
        inds[:] = -1

    else:
        # reduce dimensionality
        ep = np.median(pdist(m, metric='jaccard'))
        V, l = diffusion_map(m, ep, 'jaccard', 0, 1, 1)
        y = map_coordinates(V, l, np.arange(1, 4))

        # Create a linkage tree
        tree = linkage(y, method='ward', metric='euclidean')

        # find best clustering
        N_OF_CLUSTERS = 15
        silhs = np.zeros(shape=(N_OF_CLUSTERS,))
        silhs[0] = np.nan
        for i in np.arange(2, 16):
            c = fcluster(tree, i, 'maxclust')
            s = silhouette(y, c)
            silhs[i-1] = np.mean(s)

        # best cluster = best silhouette value = datapoints most similiar to each other
        best_clustering = np.nanargmax(silhs)

        # Repeat clustering for best silhouette
        c = fcluster(tree, best_clustering+1, 'maxclust')

        # index array for cluster names
        inds = np.zeros(shape=(c.shape[0], ), dtype=np.int32)
        inds[:] = -1

        # iterate all clusters found
        for is_ in np.unique(c):
            # find how many times each feature occurs
            # within this cluster
            c_inds = c == is_
            n_words = np.sum(m[c_inds, :], axis=0)
            if np.sum(n_words) == 0:
                print("Cluster " + str(is_) + " has no keywords.")
                continue

            # get the most common keyword
            # and use it as cluster name
            w_inds = np.argsort(n_words)
            w_inds = np.flip(w_inds, axis=0)

            inds[c_inds] = w_inds[0]

    return c, inds



from setuptools import setup
import sys

name = 'hakku'
version = '0.1.1'
description = 'Tool for article clustering based on word counts in titles and keywords.'
long_description = """
Hakku - tool for article clustering based on word counts in titles
and keywords. 

Clustering can be run from inside a script by run_files function or
from command line with python -m hakku articles1.txt articles2.txt
"""
author = 'Ilkka Pölönen, Tuomo Sipola, Paavo Nieminen, Jarno Kiesiläinen'
packages = ['hakku']
install_requires = ["numpy", "scipy", "nltk", "PyQt5", "ete3"]
zip_safe = False
package_data = {"hakku":["viridis.npy"]}
license="MIT"
python_requires = ">=3, <4"

setup(name=name,
        version=version,
        description=description,
        long_description=long_description,
        author=author,
        packages=packages,
        install_requires=install_requires,
        zip_safe=zip_safe,
        package_data=package_data,
        license=license,
        python_requires=python_requires
        )




"""
Tools for generating word occurrance matrix
"""

import sys
import re
from codecs import open as copen
import numpy as np

from nltk.stem.wordnet import WordNetLemmatizer
from nltk.stem.snowball import EnglishStemmer
from nltk.corpus import stopwords

from .article import Source

glob_lm = WordNetLemmatizer()
glob_st = EnglishStemmer()
try:
    glob_sw = stopwords.words('english')
except LookupError:
    import nltk
    nltk.download("stopwords")
    glob_sw = stopwords.words('english')

def get_words_of_string(st, extended_stopwords, stemmed_extended_stopwords):
    """Normalization, stemming, other things here."""
    words = re.split(r"\W+", st.lower())
    #words = map(glob_lm.lemmatize, words)
    words = [w for w in words if w not in glob_sw]
    words = [w for w in words if w not in extended_stopwords]
    words = map(glob_st.stem, words)
    words = [w for w in words if w not in stemmed_extended_stopwords]
    # Remove single characters and numbers:
    wordsf = filter(lambda x: len(x) > 1, words)
    wordsf = filter(lambda x: (not x.isdigit()), words)
    return list(wordsf)

def create_article_word_matrix(alist, extended_stopwords, stemmed_extended_stopwords):
    """Returns a tuple (mat, articleIDs, words) where:

         mat = binary matrix where each row corresponds to one article
               in the given article list, and each column corresponds
               to one word in the list of unique words found in titles
               and abstracts of all the articles

         articleIDs = list of article identifications (TODO: which
               format?)

         words = list of unique words found in lexicographic order
    """

    wordlist = []

    # Pass 1: get raw word list
    for a in alist:
        abstr = a.abstract.replace("All rights reserved.", "")
        wordlist = wordlist + get_words_of_string(a.title,
                extended_stopwords, stemmed_extended_stopwords)
        wordlist = wordlist + get_words_of_string(abstr,
                extended_stopwords, stemmed_extended_stopwords)

    # Make list unique:
    # "sort(unique(wordlist))":
    wordset = set(wordlist)
    orig_wlist = list(wordset)
    orig_wlist.sort()

    # # indofword and final word list comes from synonym detector:
    # (uniqlist, indofword) = detect_synonyms(orig_wlist)
    uniqlist = orig_wlist
    indofword = {}
    for ind in range(len(uniqlist)):
        indofword[uniqlist[ind]] = ind

    # Pass 2: create occurrence matrix
    nwords = len(uniqlist)
    narticles = len(alist)

    m = np.zeros((narticles, nwords))

    ai = []
    for row in range(len(alist)):
        a = alist[row]
        ai.append("%s || %s"%(a.title, a.source))

        all_words = []
        all_words = all_words + get_words_of_string(a.title,
                extended_stopwords, stemmed_extended_stopwords)
        abstr = a.abstract.replace("All rights reserved.", "")
        all_words = all_words + get_words_of_string(abstr,
                extended_stopwords, stemmed_extended_stopwords)

        for w in all_words:
            col = indofword[w]
            m[row, col] = 1

    return (m, ai, uniqlist)

def create_article_author_matrix(alist):
    """Creates a matrix articles x authors, marking with 1 the articles
    where the authors participated in.
    Input:  alist - list of articles in a data structure.
    Output: result - articles x authors binary matrix."""

    # Collect all authors to list and take unique ones.
    authors = []
    for a in alist:
        authors += a.authors
    uniq_authors = list(set(authors))

    # Which authors have contributed to which articles.
    result = map(lambda a: map(lambda author: author in a.authors, uniq_authors), alist)
    return (result, uniq_authors)

def append_from_file(l, fname, msg):
    """ Appends words from file fname to list l. Prints msg to stderr if
    file not found."""
    try:
        with copen(fname, 'r') as f:
            for line in f:
                w = line.strip()
                l.append(w)
    except IOError:
        sys.stderr.write("File %s not found. %s\n"%(fname, msg))
    return l

def ask_for_stopwords():
    print("Insert stopwords, seperate with space, emptyline to stop")
    words = []
    word = " "
    while word:
        # read words
        word = input("> ")
        if word:
            # split if user typed multiple words
            for w in word.split():
                words.append(w)

    return words

def ask_for_source(fname):
    print("Select source type for file " + str(fname))
    sources = list(Source)
    src = 0
    while src < 1 or src > len(sources):
        for s in sources:
            print(str(s.value) + ": " + str(s.name))
        try:
            src = int(input("> "))
        except ValueError:
            print("invalid number")

    return sources[src-1]

